import numpy as np
from pgmpy.models.BayesianModel import BayesianModel
from pgmpy.inference import BayesianModelProbability
from pgmpy.factors.discrete import TabularCPD


def test_bn_probability():

    # construct a tree graph structure
    model = BayesianModel([("A", "B"), ("A", "C"), ("B", "D"), ("B", "E"), ("C", "F")])
    # add CPD to each edge
    cpd_a = TabularCPD("A", 2, [[0.4], [0.6]])
    cpd_b = TabularCPD(
        "B", 3, [[0.6, 0.2], [0.3, 0.5], [0.1, 0.3]], evidence=["A"], evidence_card=[2]
    )
    cpd_c = TabularCPD(
        "C", 2, [[0.3, 0.4], [0.7, 0.6]], evidence=["A"], evidence_card=[2]
    )
    cpd_d = TabularCPD(
        "D",
        3,
        [[0.5, 0.3, 0.1], [0.4, 0.4, 0.8], [0.1, 0.3, 0.1]],
        evidence=["B"],
        evidence_card=[3],
    )
    cpd_e = TabularCPD(
        "E", 2, [[0.3, 0.5, 0.2], [0.7, 0.5, 0.8]], evidence=["B"], evidence_card=[3]
    )
    cpd_f = TabularCPD(
        "F", 3, [[0.3, 0.6], [0.5, 0.2], [0.2, 0.2]], evidence=["C"], evidence_card=[2]
    )
    model.add_cpds(cpd_a, cpd_b, cpd_c, cpd_d, cpd_e, cpd_f)

    """ transposed conditional probabilities
    C {(0,): array([0.3, 0.7]), (1,): array([0.4, 0.6])}
    F {(0,): array([0.3, 0.5, 0.2]), (1,): array([0.6, 0.2, 0.2])}
    B {(0,): array([0.6, 0.3, 0.1]), (1,): array([0.2, 0.5, 0.3])}
    E {(0,): array([0.3, 0.7]), (1,): array([0.5, 0.5]), (2,): array([0.2, 0.8])}
    D {(0,): array([0.5, 0.4, 0.1]), (1,): array([0.3, 0.4, 0.3]), (2,): array([0.1, 0.8, 0.1])}
    """

    inference = BayesianModelProbability(model)
    ordering = ["A", "C", "F", "B", "E", "D"]

    x0 = np.array([[0, 0, 0, 0, 0, 0]])
    x1 = np.array([[1, 1, 1, 1, 1, 1]])
    x2 = np.array([[1, 1, 2, 2, 1, 2]])
    X = np.concatenate([x0, x1, x2], axis=0)

    p0 = 0.4 * 0.3 * 0.3 * 0.6 * 0.3 * 0.5
    p1 = 0.6 * 0.6 * 0.2 * 0.5 * 0.5 * 0.4
    p2 = 0.6 * 0.6 * 0.2 * 0.3 * 0.8 * 0.1

    logp = inference.log_probability(x0, ordering)
    p = np.exp(logp)
    np.testing.assert_almost_equal(p[0], p0)

    logp = inference.log_probability(x1, ordering)
    p = np.exp(logp)
    np.testing.assert_almost_equal(p[0], p1)

    logp = inference.log_probability(x2, ordering)
    p = np.exp(logp)
    np.testing.assert_almost_equal(p[0], p2)

    logp = inference.log_probability(X, ordering)
    p = np.exp(logp)
    np.testing.assert_array_almost_equal(p, [p0, p1, p2])

    data = np.array(
        [
            [1, 0, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 2],
            [1, 0, 0, 1, 0, 1],
            [0, 1, 2, 0, 1, 1],
            [1, 1, 1, 0, 1, 0],
            [1, 1, 2, 1, 1, 0],
            [0, 1, 0, 0, 1, 0],
            [0, 0, 2, 0, 0, 1],
            [0, 1, 0, 0, 1, 2],
            [1, 0, 1, 2, 0, 1],
        ]
    )
    logp = inference.log_probability(data, ordering)
    p = np.exp(logp)

    p_check = np.array(
        [
            0.012,
            0.0054,
            0.0072,
            0.009408,
            0.00504,
            0.0054,
            0.03528,
            0.001728,
            0.007056,
            0.00576,
        ]
    )
    np.testing.assert_array_almost_equal(p, p_check)

    llsum = inference.score(data, ordering)
    np.testing.assert_almost_equal(llsum, -49.57170403869862)

    # use topological ordering of model to interpret data
    logp = inference.log_probability(data)
    p = np.exp(logp)
    np.testing.assert_array_almost_equal(p, p_check)

    llsum = inference.score(data)
    np.testing.assert_almost_equal(llsum, -49.57170403869862)



from .base import Inference
from .ExactInference import BeliefPropagation
from .ExactInference import VariableElimination
from .dbn_inference import DBNInference
from .bn_inference import BayesianModelInference, BayesianModelProbability
from .mplp import Mplp

__all__ = [
    "Inference",
    "VariableElimination",
    "DBNInference",
    "BayesianModelInference",
    "BayesianModelProbability",
    "BeliefPropagation",
    "BayesianModelSampling",
    "GibbsSampling",
    "Mplp",
    "continuous",
]



#!/usr/bin/env python3

import networkx as nx

class MG(nx.MultiDiGraph):
    """
    Class representing a Mixed Graph. A mixed graph can contain both a directed edge and
    a bi-directed edge. Bi-directed edges are represented using two edges between the same
    nodes in opposite directions.
    """
    def __init__(self, ebunch=None):
        super(MG, self).__init__(ebunch)

    def add_node(self, node, weight=None):
        """
        Adds a single node to the Graph.

        Parameters
        ----------
        node: str, int, or any hashable python object.
            The node to add to the graph.

        weight: int, float
            The weight of the node.

        Examples
        --------
        >>> from pgmpy.base import MG
        >>> G = MG()
        >>> G.add_node(node='A')
        >>> sorted(G.nodes())
        ['A']

        Adding a node with some weight.
        >>> G.add_node(node='B', weight=0.3)

        The weight of these nodes can be accessed as:
        >>> G.nodes['B']
        {'weight': 0.3}
        >>> G.nodes['A']
        {'weight': None}
        """
        super(MG, self).add_node(node, weight=weight)

    def add_nodes_from(self, nodes, weigths=None):
        """
        Add multiple nodes to the Graph.

        **The behviour of adding weights is different than in networkx.

        Parameters
        ----------
        nodes: iterable container
            A container of nodes (list, dict, set, or any hashable python
            object).

        weights: list, tuple (default=None)
            A container of weights (int, float). The weight value at index i
            is associated with the variable at index i.

        Examples
        --------
        >>> from pgmpy.base import DAG
        >>> G = MG()
        >>> G.add_nodes_from(nodes=['A', 'B', 'C'])
        >>> G.nodes()
        NodeView(('A', 'B', 'C'))

        Adding nodes with weights:
        >>> G.add_nodes_from(nodes=['D', 'E'], weights=[0.3, 0.6])
        >>> G.nodes['D']
        {'weight': 0.3}
        >>> G.nodes['E']
        {'weight': 0.6}
        >>> G.nodes['A']
        {'weight': None}
        """
        nodes = list(nodes)

        if weights:
            if len(nodes) != len(weights):
                raise ValueError(
                    "The number of elements in nodes and weights" "should be equal."
                )
            for index in range(len(nodes)):
                self.add_node(node=nodes[index], weight=weights[index])
        else:
            for node in nodes:
                self.add_node(node=node)

    def add_edge(self, u, v, weight=None):
        """
        Add an edge between u and v.

        The nodes u and v will be automatically added if they are
        not already in the graph.

        Parameters
        ----------
        u, v : nodes
            Nodes can be any hashable Python object.

        weight: int, float (default=None)
            The weight of the edge

        Examples
        --------
        >>> from pgmpy.base import MG
        >>> G = MG()
        >>> G.add_nodes_from(nodes=['Alice', 'Bob', 'Charles'])
        >>> G.add_edge(u='Alice', v='Bob')
        >>> G.nodes()
        NodeView(('Alice', 'Bob', 'Charles'))
        >>> G.edges()
        OutEdgeView([('Alice', 'Bob')])

        When the node is not already present in the graph:
        >>> G.add_edge(u='Alice', v='Ankur')
        >>> G.nodes()
        NodeView(('Alice', 'Ankur', 'Bob', 'Charles'))
        >>> G.edges()
        OutEdgeView([('Alice', 'Bob'), ('Alice', 'Ankur')])

        Adding edges with weight:
        >>> G.add_edge('Ankur', 'Maria', weight=0.1)
        >>> G.edge['Ankur']['Maria']
        {'weight': 0.1}
        """
        super(DAG, self).add_edge(u, v, weight=weight)

    def add_edges_from(self, ebunch, weights=None):
        """
        Add all the edges in ebunch.

        If nodes referred in the ebunch are not already present, they
        will be automatically added. Node names can be any hashable python
        object.

        **The behavior of adding weights is different than networkx.

        Parameters
        ----------
        ebunch : container of edges
            Each edge given in the container will be added to the graph.
            The edges must be given as 2-tuples (u, v).

        weights: list, tuple (default=None)
            A container of weights (int, float). The weight value at index i
            is associated with the edge at index i.

        Examples
        --------
        >>> from pgmpy.base import MG
        >>> G = MG()
        >>> G.add_nodes_from(nodes=['Alice', 'Bob', 'Charles'])
        >>> G.add_edges_from(ebunch=[('Alice', 'Bob'), ('Bob', 'Charles')])
        >>> G.nodes()
        NodeView(('Alice', 'Bob', 'Charles'))
        >>> G.edges()
        OutEdgeView([('Alice', 'Bob'), ('Bob', 'Charles')])

        When the node is not already in the model:
        >>> G.add_edges_from(ebunch=[('Alice', 'Ankur')])
        >>> G.nodes()
        NodeView(('Alice', 'Bob', 'Charles', 'Ankur'))
        >>> G.edges()
        OutEdgeView([('Alice', 'Bob'), ('Bob', 'Charles'), ('Alice', 'Ankur')])

        Adding edges with weights:
        >>> G.add_edges_from([('Ankur', 'Maria'), ('Maria', 'Mason')],
        ...                  weights=[0.3, 0.5])
        >>> G.edge['Ankur']['Maria']
        {'weight': 0.3}
        >>> G.edge['Maria']['Mason']
        {'weight': 0.5}
        """
        ebunch = list(ebunch)

        if weights:
            if len(ebunch) != len(weights):
                raise ValueError(
                    "The number of elements in ebunch and weights" "should be equal"
                )
            for index in range(len(ebunch)):
                self.add_edge(ebunch[index][0], ebunch[index][1], weight=weights[index])
        else:
            for edge in ebunch:
                self.add_edge(edge[0], edge[1])

class MAG(nx.MultiDiGraph):
    """
    Class representing Maximal Ancestral Graph (MAG)[1].

    References
    ----------
    [1] Zhang, Jiji. "Causal reasoning with ancestral graphs." Journal of Machine Learning Research 9 (2008): 1437-1474.
    """
    def __init__(self, ebunch=None):
        pass
